"""Connection stuffs."""

import ssl
from asyncio import StreamReader, StreamWriter, open_connection
from ssl import SSLContext
from typing import Dict, Optional
from urllib.parse import ParseResult

import h2.config
import h2.connection
import h2.events

from aiosonic.connectors import TCPConnector
from aiosonic.exceptions import (
    HttpParsingError,
    MissingReaderException,
    MissingWriterException,
)
from aiosonic.http2 import Http2Handler
from aiosonic.tcp_helpers import keepalive_flags
from aiosonic.types import ParsedBodyType


class Connection:
    """Connection class.

    This class represents a connection to a remote server, managing the communication
    through a socket. It is designed to handle both HTTP/1.1 and HTTP/2 protocols.

    Attributes:
        connector (TCPConnector): An instance of the TCPConnector class responsible
            for managing the connection pool.
        reader (Optional[StreamReader]): A StreamReader for reading data from the socket.
        writer (Optional[StreamWriter]): A StreamWriter for efficiently writing data
            to the socket.
        keep (bool): A flag indicating whether the connection should be kept alive.
        key (Optional[str]): A key identifying the connection based on the hostname and port.
        blocked (bool): A flag indicating whether the connection is currently blocked,
            meaning it is in use and should not be reinserted into the pool until all
            data has been read.
        temp_key (Optional[str]): A temporary key used during the connection setup process.
        requests_count (int): The count of requests made over the connection.
        h2conn (Optional[h2.connection.H2Connection]): An instance of the H2Connection
            class representing the HTTP/2 connection.
        h2handler (Optional[Http2Handler]): An instance of the Http2Handler class
            responsible for handling HTTP/2 requests.
    """

    def __init__(self, connector: TCPConnector) -> None:
        self.connector = connector
        self.reader: Optional[StreamReader] = None
        self.writer: Optional[StreamWriter] = None

        self.keep = False  # keep alive flag
        self.key = None
        self.blocked = False
        self.temp_key: Optional[str] = None
        self.requests_count = 0
        self.background_tasks = set()

        self.h2conn: Optional[h2.connection.H2Connection] = None
        self.h2handler: Optional[Http2Handler] = None

        self._verify = True
        self.proxy_connected = False

    @property
    def is_connected(self):
        return not self.writer is None

    async def connect(
        self,
        urlparsed: ParseResult,
        dns_info: dict,
        verify: bool,
        ssl_context: SSLContext,
        http2: bool = False,
    ) -> None:
        """Connect with timeout."""
        self._verify = verify
        await self._connect(urlparsed, verify, ssl_context, dns_info, http2)

    def write(self, data: bytes):
        """Write data in the socket."""
        if not self.writer:
            raise MissingWriterException("writer not set.")
        self.writer.write(data)

    async def readline(self):
        """Read data until line break"""
        if not self.reader:
            raise MissingReaderException("reader not set.")
        return await self.reader.readline()

    async def readexactly(self, size: int):
        """Read exactly size of bytes"""
        if not self.reader:
            raise MissingReaderException("reader not set.")
        return await self.reader.readexactly(size)

    async def read(self, size: int = -1):
        """Read up to size of bytes"""
        if not self.reader:
            raise MissingReaderException("reader not set.")
        return await self.reader.read(size)

    async def readuntil(self, separator: bytes = b"\n"):
        """Read until separator"""
        if not self.reader:
            raise MissingReaderException("reader not set.")
        return await self.reader.readuntil(separator)

    async def _connect(
        self,
        urlparsed: ParseResult,
        verify: bool,
        ssl_context: SSLContext,
        dns_info,
        http2: bool,
    ) -> None:
        """Get reader and writer."""
        if not urlparsed.hostname:
            raise HttpParsingError("missing hostname")

        key = f"{urlparsed.hostname}-{urlparsed.port}"

        def is_closing():
            return True  # noqa

        if self.writer:
            is_closing = self.writer.is_closing  # type: ignore

        dns_info_copy = dns_info.copy()
        dns_info_copy["server_hostname"] = dns_info_copy.pop("hostname")
        dns_info_copy["flags"] = dns_info_copy["flags"] | keepalive_flags()

        if not (self.key and key == self.key and not is_closing() and
                self.requests_count <= self.connector.conn_max_requests
            ):
            self.close()

            if urlparsed.scheme == "https":
                ssl_context = ssl_context or get_default_ssl_context(verify, http2)
            else:
                del dns_info_copy["server_hostname"]
            port = urlparsed.port or (443 if urlparsed.scheme == "https" else 80)
            dns_info_copy["port"] = port

            self.reader, self.writer = await open_connection(
                **dns_info_copy, ssl=ssl_context
            )

            self.temp_key = key
            await self._connection_made()

    async def _connection_made(self) -> None:
        tls_conn = self.writer.get_extra_info("ssl_object")
        if not tls_conn:
            return

        # Always prefer the result from ALPN to that from NPN.
        negotiated_protocol = tls_conn.selected_alpn_protocol()
        if negotiated_protocol is None:
            negotiated_protocol = tls_conn.selected_npn_protocol()

        if negotiated_protocol == "h2":  # pragma: no cover
            config = h2.config.H2Configuration()
            self.h2conn = h2.connection.H2Connection(config=config)
            self.h2handler = Http2Handler(self)

    def keep_alive(self) -> None:
        """Check if keep alive."""
        self.keep = True

    def block_until_read_chunks(self):
        """Check if keep alive."""
        self.blocked = True

    def release(self) -> None:
        """Release connection."""
        self.requests_count += 1
        # ensure unblock conn object after read
        self.blocked = False
        self.connector.release(self)

    def ensure_released(self):
        """Ensure the connection is released."""
        if self.blocked:
            self.release()

    def close(self) -> None:
        """Close connection if opened."""
        if self.reader:
            try:
                self.writer._transport.abort()
            except:
                pass

            self.reader, self.writer = None, None
        self.proxy_connected = False

    async def upgrade(self, ssl_context: SSLContext = None):
        ssl_context = ssl_context or get_default_ssl_context(self._verify)
        if not self.writer:
            raise MissingWriterException()
        await self.writer.start_tls(ssl_context)

    async def http2_request(
        self, headers: Dict[str, str], body: Optional[ParsedBodyType]
    ):
        if self.h2handler:  # pragma: no cover
            return await self.h2handler.request(headers, body)

    async def __aenter__(self):
        """Get connection from pool."""
        return self

    async def __aexit__(self, exc_type: None, exc: None, tb: None) -> None:
        """Release connection."""
        if self.keep and not exc:
            self.key = self.temp_key
        else:
            self.key = None
            self.h2conn = None

        if not self.blocked:
            self.release()
            if self.h2handler:  # pragma: no cover
                self.h2handler.cleanup()


def get_default_ssl_context(verify=True, http2=False):
    if http2:  # pragma: no cover
        ssl_context = _get_http2_ssl_context()
    else:
        ssl_context = ssl.create_default_context(
            ssl.Purpose.SERVER_AUTH,
        )

    if not verify:
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE
    return ssl_context


def _get_http2_ssl_context():
    """
    This function creates an SSLContext object that is suitably configured for
    HTTP/2. If you're working with Python TLS directly, you'll want to do the
    exact same setup as this function does.
    """
    # Get the basic context from the standard library.
    ctx = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)

    # RFC 7540 Section 9.2: Implementations of HTTP/2 MUST use TLS version 1.2
    # or higher. Disable TLS 1.1 and lower.
    ctx.options |= (
        ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 | ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
    )

    # RFC 7540 Section 9.2.1: A deployment of HTTP/2 over TLS 1.2 MUST disable
    # compression.
    ctx.options |= ssl.OP_NO_COMPRESSION

    # RFC 7540 Section 9.2.2: "deployments of HTTP/2 that use TLS 1.2 MUST
    # support TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256". In practice, the
    # blocklist defined in this section allows only the AES GCM and ChaCha20
    # cipher suites with ephemeral key negotiation.
    ctx.set_ciphers("ECDHE+AESGCM:ECDHE+CHACHA20:DHE+AESGCM:DHE+CHACHA20")

    # We want to negotiate using NPN and ALPN. ALPN is mandatory, but NPN may
    # be absent, so allow that. This setup allows for negotiation of HTTP/1.1.
    ctx.set_alpn_protocols(["h2", "http/1.1"])

    try:
        if hasattr(ctx, "_set_npn_protocols"):
            ctx.set_npn_protocols(["h2", "http/1.1"])
    except NotImplementedError:
        pass

    return ctx



import os
from io import IOBase
from random import randint
from typing import Optional, Union

from aiosonic.resolver import get_loop

RANDOM_RANGE = (1000, 9999)
_CHUNK_SIZE = 1024 * 1024  # 1mb


class MultipartForm:
    """
    A class to handle multipart form data for HTTP requests.

    Example:

    .. code-block:: python

        import asyncio
        import aiosonic
        from multipart import MultipartForm
        
        async def upload_file():
            client = aiosonic.HTTPClient()
            form = MultipartForm()
        
            # Add a text field
            form.add_field("field1", "value1")
        
            # Add a file to upload
            form.add_file("file1", "path/to/your/file.txt")
        
            # Make the POST request with MultipartForm directly
            url = "https://your-upload-endpoint.com/upload"
            response = await client.post(url, data=form)
        
            print("Response Status:", response.status_code)
            response_data = await response.text()
            print("Response Body:", response_data)
        
        if __name__ == '__main__':
            asyncio.run(upload_file())
    """

    def __init__(self):
        """Initializes an empty list for fields and generates a boundary."""
        self.fields = []
        self.boundary = f"boundary-{randint(*RANDOM_RANGE)}"

    def add_field(
        self, name: str, value: Union[str, IOBase], filename: Optional[str] = None
    ):
        """Adds a field to the multipart form data.

        Args:
            name (str): The name of the field.
            value (Union[str, IOBase]): The value of the field. Can be a string or a file-like object.
            filename (Optional[str]): The name of the file, if the value is a file-like object. 
                                      Defaults to the file's name if not provided.
        """
        if isinstance(value, IOBase):
            if not filename:
                filename = os.path.basename(value.name)  # Default to the file's name
            self.fields.append((name, value, filename))
        else:
            self.fields.append((name, value))

    def add_file(self, name: str, file_path: str, filename: Optional[str] = None):
        """Adds a file to the multipart form data.

        The file is opened and it is closed after the request is sent.

        Args:
            name (str): The name of the file field.
            file_path (str): The file path of the file to be added.
            filename (Optional[str]): The name of the file for the target server. defaults to the file's name.
        """
        file = open(file_path, "rb")
        self.add_field(name, file, filename or os.path.basename(file_path))

    async def _read_file(self, file_obj: IOBase):
        loop = get_loop()
        while True:
            data = await loop.run_in_executor(None, file_obj.read, _CHUNK_SIZE)
            if not data:
                break
            yield data

    async def _generate_chunks(self):
        """Yields chunks of the multipart buffer containing all fields asynchronously."""
        for field in self.fields:
            yield (f"--{self.boundary}\r\n").encode()
            if isinstance(field[1], IOBase):
                to_write = (
                    "Content-Disposition: form-data; "
                    + f'name="{field[0]}"; filename="{field[2]}"\r\n\r\n'
                )
                yield to_write.encode()

                # Read the file asynchronously
                async for data in self._read_file(field[1]):
                    yield data
                field[1].close()
            else:
                yield (
                    f'Content-Disposition: form-data; name="{field[0]}"\r\n\r\n'
                ).encode()
                yield field[1].encode() + b"\r\n"

        yield (f"--{self.boundary}--").encode()

    async def get_buffer(self):
        """Returns an asynchronous iterator that generates the constructed multipart buffer."""
        async for chunk in self._generate_chunks():
            yield chunk

    async def get_body_size(self):
        """Calculates the total size of the multipart body and returns it along with the body itself.

        This function asynchronously constructs the body by iterating over the chunks
        generated by the get_buffer method. It accumulates the total size of the body
        in bytes while building the complete body as a byte string.

        Returns:
            tuple: A tuple containing the complete multipart body as bytes and its size in bytes.
        """
        body = b""
        size = 0
        async for chunk in self.get_buffer():
            body += chunk
            size += len(chunk)
        return body, size

    def get_headers(self, size=None):
        """Returns the headers for the multipart form data."""
        headers = {"Content-Type": f"multipart/form-data; boundary={self.boundary}"}
        if size:
            headers["Content-Length"] = str(size)
        return headers
